#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Analysis;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Problems.Instances;


namespace HeuristicLab.Problems.GeneticProgramming.GlucosePrediction {
  [Item("Blood Glucose Forecast", "See MedGEC Workshop at GECCO 2016")]
  [Creatable(CreatableAttribute.Categories.GeneticProgrammingProblems, Priority = 999)]
  [StorableClass]
  public sealed class Problem : SymbolicExpressionTreeProblem, IRegressionProblem, IProblemInstanceConsumer<IRegressionProblemData>, IProblemInstanceExporter<IRegressionProblemData> {

    #region parameter names
    private const string ProblemDataParameterName = "ProblemData";
    #endregion

    #region Parameter Properties
    IParameter IDataAnalysisProblem.ProblemDataParameter { get { return ProblemDataParameter; } }

    public IValueParameter<IRegressionProblemData> ProblemDataParameter {
      get { return (IValueParameter<IRegressionProblemData>)Parameters[ProblemDataParameterName]; }
    }
    #endregion

    #region Properties
    public IRegressionProblemData ProblemData {
      get { return ProblemDataParameter.Value; }
      set { ProblemDataParameter.Value = value; }
    }
    IDataAnalysisProblemData IDataAnalysisProblem.ProblemData { get { return ProblemData; } }
    #endregion

    public event EventHandler ProblemDataChanged;

    public override bool Maximization {
      get { return true; }
    }

    #region item cloning and persistence
    // persistence
    [StorableConstructor]
    private Problem(bool deserializing) : base(deserializing) { }
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      RegisterEventHandlers();
    }

    // cloning 
    private Problem(Problem original, Cloner cloner)
      : base(original, cloner) {
      RegisterEventHandlers();
    }
    public override IDeepCloneable Clone(Cloner cloner) { return new Problem(this, cloner); }
    #endregion

    public Problem()
      : base() {
      Parameters.Add(new ValueParameter<IRegressionProblemData>(ProblemDataParameterName, "The data for the glucose prediction problem", new RegressionProblemData()));

      var g = new SimpleSymbolicExpressionGrammar(); // empty grammar is replaced in UpdateGrammar()
      base.Encoding = new SymbolicExpressionTreeEncoding(g, 100, 17);

      UpdateGrammar();
      RegisterEventHandlers();
    }


    public override double Evaluate(ISymbolicExpressionTree tree, IRandom random) {
      var problemData = ProblemData;
      var rows = problemData.TrainingIndices.ToArray();
      var target = problemData.Dataset.GetDoubleValues(problemData.TargetVariable, rows);
      var predicted0 = Interpreter.Apply(tree.Root.GetSubtree(0).GetSubtree(0).GetSubtree(0), problemData.Dataset, rows);
      var predicted1 = Interpreter.Apply(tree.Root.GetSubtree(0).GetSubtree(0).GetSubtree(1), problemData.Dataset, rows);
      var predicted2 = Interpreter.Apply(tree.Root.GetSubtree(0).GetSubtree(0).GetSubtree(2), problemData.Dataset, rows);

      var pred0_rsq = Rsq(predicted0, target);
      var pred1_rsq = Rsq(predicted1, target);
      var pred2_rsq = Rsq(predicted2, target);
      return pred0_rsq + pred1_rsq + pred2_rsq;
    }

    private double Rsq(IEnumerable<double> predicted, IEnumerable<double> target) {
      // only take predictions for which the target is not NaN
      var selectedTuples = target.Zip(predicted, Tuple.Create).Where(t => !double.IsNaN(t.Item1)).ToArray();
      target = selectedTuples.Select(t => t.Item1);
      predicted = selectedTuples.Select(t => t.Item2);

      OnlineCalculatorError errorState;
      var r = OnlinePearsonsRCalculator.Calculate(target, predicted, out errorState);
      if (errorState != OnlineCalculatorError.None) r = 0;
      return r * r;
    }

    public override void Analyze(ISymbolicExpressionTree[] trees, double[] qualities, ResultCollection results,
      IRandom random) {
      base.Analyze(trees, qualities, results, random);

      if (!results.ContainsKey("Solution")) {
        results.Add(new Result("Solution", typeof(IRegressionSolution)));
      }
      if (!results.ContainsKey("Terms")) {
        results.Add(new Result("Terms", typeof(DataTable)));
      }

      var bestTree = trees.First();
      var bestQuality = qualities.First();
      for (int i = 1; i < trees.Length; i++) {
        if (qualities[i] > bestQuality) {
          bestQuality = qualities[i];
          bestTree = trees[i];
        }
      }


      var clonedProblemData = (IRegressionProblemData)ProblemData.Clone();
      var rows = clonedProblemData.TrainingIndices.ToArray();
      var target = clonedProblemData.Dataset.GetDoubleValues(clonedProblemData.TargetVariable, rows).ToArray();
      var predicted0 =
        Interpreter.Apply(bestTree.Root.GetSubtree(0).GetSubtree(0).GetSubtree(0), clonedProblemData.Dataset, rows)
          .ToArray();
      var predicted1 =
        Interpreter.Apply(bestTree.Root.GetSubtree(0).GetSubtree(0).GetSubtree(1), clonedProblemData.Dataset, rows)
          .ToArray();
      var predicted2 =
        Interpreter.Apply(bestTree.Root.GetSubtree(0).GetSubtree(0).GetSubtree(2), clonedProblemData.Dataset, rows)
          .ToArray();

      var termsTable = new HeuristicLab.Analysis.DataTable("Terms");
      var r0 = new DataRow("GlucTerm", "GlucTerm", predicted0);
      var r1 = new DataRow("InsTerm", "InsTerm", predicted1);
      r1.VisualProperties.SecondYAxis = true;
      var r2 = new DataRow("ChTerm", "ChTerm", predicted2);
      r2.VisualProperties.SecondYAxis = true;
      var r3 = new DataRow("Target", "Target", target);
      termsTable.Rows.Add(r0);
      termsTable.Rows.Add(r1);
      termsTable.Rows.Add(r2);
      termsTable.Rows.Add(r3);
      results["Terms"].Value = termsTable;


      var filteredPredicted0 = rows.Where(r => !double.IsNaN(target[r])).Select(r => predicted0[r]).ToArray();
      var filteredPredicted1 = rows.Where(r => !double.IsNaN(target[r])).Select(r => predicted1[r]).ToArray();
      var filteredPredicted2 = rows.Where(r => !double.IsNaN(target[r])).Select(r => predicted2[r]).ToArray();
      var filteredTarget = target.Where(t => !double.IsNaN(t)).ToArray();


      var ds = new ModifiableDataset(new string[] { "pred0", "pred1", "pred2", "target" },
        new List<IList>
        {
          filteredPredicted0.ToList(),
          filteredPredicted1.ToList(),
          filteredPredicted2.ToList(),
          filteredTarget.ToList()
        });
      var lrProbData = new RegressionProblemData(ds, new string[] { "pred0", "pred1", "pred2" }, "target");
      lrProbData.TrainingPartition.Start = clonedProblemData.TrainingPartition.Start;
      lrProbData.TrainingPartition.End = clonedProblemData.TrainingPartition.End;
      lrProbData.TestPartition.Start = clonedProblemData.TestPartition.Start;
      lrProbData.TestPartition.End = clonedProblemData.TestPartition.End;

      try {
        double rmsError, cvRmsError;
        var lrSolution = HeuristicLab.Algorithms.DataAnalysis.LinearRegression.CreateLinearRegressionSolution(
          lrProbData, out rmsError, out cvRmsError);
        results["Solution"].Value = lrSolution;
      } catch (Exception) {
        // ignore
      }
    }

    #region events
    private void RegisterEventHandlers() {
      ProblemDataParameter.ValueChanged += new EventHandler(ProblemDataParameter_ValueChanged);
      if (ProblemDataParameter.Value != null) ProblemDataParameter.Value.Changed += new EventHandler(ProblemData_Changed);
    }

    private void ProblemDataParameter_ValueChanged(object sender, EventArgs e) {
      ProblemDataParameter.Value.Changed += new EventHandler(ProblemData_Changed);
      OnProblemDataChanged();
      OnReset();
    }

    private void ProblemData_Changed(object sender, EventArgs e) {
      OnReset();
    }

    private void OnProblemDataChanged() {
      UpdateGrammar();

      var handler = ProblemDataChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    private void UpdateGrammar() {
      // whenever ProblemData is changed we create a new grammar with the necessary symbols
      var g = new Grammar();
      Encoding.Grammar = g;
    }
    #endregion

    #region Import & Export
    public void Load(IRegressionProblemData data) {
      Name = data.Name;
      Description = data.Description;
      ProblemData = data;
    }

    public IRegressionProblemData Export() {
      return ProblemData;
    }
    #endregion
  }
}
